"plot.tree" <-
function(t,horizontal=F,labels=F,cex.lab=0.8,timeaxis=F,maxtime,labcols="black",linecol="black",...) 
{
    
    if (labels==F) mr <- c(1,1,0,0)
    else mr <- c(2,1,0,2);
    if (timeaxis) mr <- mr+c(0,0,0,3);
    opar <- par(mar=mr,mgp=c(2,0.5,0));
    if (horizontal) xl <- c(-1,0.15)
    else xl <- c(0,1.05);
    plot(2,2,axes=F,xlab="",ylab="",xaxt="n",
         yaxt="n",xlim=xl,ylim=c(-0.05,1));
    if (length(labels)<2) {
	if (labels==T)  labs <- t$labels
	else labs <- NULL;
    } else labs <- labels;
    if (missing(maxtime)) timescale <- 1/max(t$t)
    else timescale <- 1/maxtime
    
    n <- length(t$t)+1;
    a <- plottriple(t,n,2*n-1,0,xscale=1/(n-1),yscale=timescale
                    ,labs,horizontal,cex.lab,labcols=labcols,linecol=linecol,...);
    
    if (timeaxis) {
	if (maxtime<0.8) pts <- seq(0,maxtime,0.1)
	else if (maxtime<1.4) pts <- seq(0,maxtime,0.2)
        else if (maxtime>4) pts <- seq(0,maxtime,1)
	else  pts <- seq(0,maxtime,0.5);
	posit <- pts/maxtime;
	if (horizontal) {
	    axis(1,at=-posit,labels=pts,cex=0.8);
	    mtext("Scaled Time",1,line=1.5,cex=0.8);
	} else {
	    axis(4,at=posit,labels=pts,cex=0.8);
	    mtext("Scaled Time",4,line=1.6,cex=0.8);
	}
    }
    par(opar);
}

"plottriple" <-
  function(t,n,where,xpos,xscale,yscale,labels,reverse=F,cex.lab=12/n,labcols="black",linecol="black",...) 
{
###   are we at a leaf ?  If so then we may need to plot a label
    if (where <= n) { 
	if (length(labels)>0) {
	    if (reverse)
              text(0.01,xpos,labels[where],cex=cex.lab,adj=c(0,0.5),col=labcols,...)
	    else 
              text(xpos,-0.01,labels[where],cex=cex.lab,adj=c(0.5,1),col=labcols,...);
	}
	return(list(x=xpos,y=0,xpos=xpos+xscale));
    } else {
###   plot the tree to the left
        left <- plottriple(t,n,t$d1[where-n],xpos,xscale,yscale,labels
                           ,reverse,cex.lab,labcols,linecol,...);
###   plot the tree to the right
        right <- plottriple(t,n,t$d2[where-n],left$xpos,xscale,yscale,labels
                          ,reverse,cex.lab,labcols,linecol,...);
        y1 <- left$y;y2 <- right$y;
        x1 <- left$x;x2 <- right$x;
        yt <- t$t[where-n]*yscale;
        if (reverse) 
          segments(-c(y1,yt,yt),c(x1,x1,x2),-c(yt,yt,y2),c(x1,x2,x2),col=linecol,...)
        else segments(c(x1,x1,x2),c(y1,yt,yt),c(x1,x2,x2),c(yt,yt,y2),col=linecol,...);
        return(list(x=(x1+x2)/2,y=yt,xpos=right$xpos));
    }
}
