#include <R.h>
#include <R_ext/PrtUtil.h>
#include <Rmath.h>
#include "treeutil.h"

#ifdef WIN32
#define DECLSPEC __declspec(dllexport) 
#else 
#define DECLSPEC 
#endif

typedef struct nodetype {
  int *haplotype;	        // contains active/non-active
                                // information or haplotypes
  struct nodetype *desc_left;	// if a coalescent node
  struct nodetype *desc_right;	// these are desc_left and right
  int recomb_pos;		// position of recombination
  double time;
  int leaf;
} node;

typedef struct recombtreetype {
  node *sample;	                // the sample
  node **ancestor;              // pointers to ancestral nodes
  node **root;	                // vector of pointers to roots for different positions
  int n,loci,nancs;             // sample size number of loci, number of ancestral nodes
} recombtree;
/*************************************************************************/
DECLSPEC
void simulatelinkedSTR(int *n, int *loci, double *r,int *nr
			    ,double *theta, int *data);
DECLSPEC
void simulatetimeroot(int *n, int *loci, double *r, int *nr, double *troot);
DECLSPEC
void simulatelinkedSNP(int *n, int *loci, double *r,int *nr, int *data);
node *node_vector(int n, int loci);
node *getnewnode(int loci);
void initial_rates(int n, int loci, double *r,double rate[2]);
double *get_recombrates(node **current,int n, int loci, double *r);
int which_recomb(node *any, int loci, double *r);
double diffrrate(int *h1, int *h2, int *h3,int loci, double *r);
recombtree generaterecombtree(int n, int loci, double *r, int nr);
void destroy_recombtree(recombtree *t);
node *findcoalnode(node *any);
/*************************************************************************/
void recurse_mutSTR(node *any, int l,double tabove, int STR, double theta)
{
  node *next;
  next=findcoalnode(any);
  
  next->haplotype[l]=STR+(int)rpois(theta*(tabove-next->time)*0.25) -
    (int)rpois(theta*(tabove-next->time)*0.25);
  if (next->desc_left!=NULL) {
    recurse_mutSTR(next->desc_left,l,next->time,next->haplotype[l],theta);
    recurse_mutSTR(next->desc_right,l,next->time,next->haplotype[l],theta);
  }
}
/*************************************************************************/
void mutaterecombSTR(recombtree *rt, double *theta)
{
  int locus;

  for (locus=0;locus<rt->loci;locus++) {
    rt->root[locus]->haplotype[locus]=0;
    recurse_mutSTR(rt->root[locus]->desc_left,locus
		   ,rt->root[locus]->time,0,theta[locus]);
    recurse_mutSTR(rt->root[locus]->desc_right,locus
		   ,rt->root[locus]->time,0,theta[locus]);
  }
}
/*************************************************************************/
double recomb_tree_length(node *any,double tabove) 
{
  double l_l,l_r;
  node *next;
  next=findcoalnode(any);
 
  if (next->desc_left!=NULL) {
    l_l=recomb_tree_length(next->desc_left,next->time);
    l_r = recomb_tree_length(next->desc_right,next->time);
    return tabove-next->time+l_l+l_r;
  } else return tabove-next->time;
}
/*************************************************************************/
node *findcoalnode(node *any)
{
  node *here;
  here=any;
  for (;;) {
    if (here->desc_left==NULL) return here;  // found a leaf
    else if (here->desc_right==NULL) here=here->desc_left;
    else return here;
  }
  return here;
 }
/*************************************************************************/
int recursecount_leaves(node *any)
{  
  node *next;
  next=findcoalnode(any);
  if (next->desc_left==NULL) {
    if (next->leaf!=0) Rprintf("leaf %d\n",next->leaf);
    return 1;
  }

  else return  recursecount_leaves(next->desc_left)+recursecount_leaves(next->desc_right);
}
/*************************************************************************/
void countleaves(recombtree *rt)
{
  int i;
  for (i=0;i<rt->loci;i++) 
    Rprintf("locus %d  leaves %d \n",i+1,recursecount_leaves(rt->root[i]));
}
/*************************************************************************/
void mutatebelow(node *any, int locus,int what)
{
  node *next;
  next=findcoalnode(any);
  next->haplotype[locus]=what;
  if (next->desc_left==NULL&&what==1) Rprintf("changing leaf\n");
  if (next->desc_left!=NULL) {
    mutatebelow(next->desc_left,locus,what);
    mutatebelow(next->desc_right,locus,what);
  }
}
/*************************************************************************/
int recursemutaterecombsnp(node *any,double *muttime,double tabove, 
			   int locus) 
{
  int mut;
  node *next;
 
  next=findcoalnode(any);

  if (*muttime<0.0) Rprintf("error aaaaaaaa\n");
  *muttime-=tabove-next->time;
  if (*muttime<0.0) {        // mutation above this point
    Rprintf("mutation\n");
    next->haplotype[locus]=1; // set equal to 1, and all below the same
    mutatebelow(next,locus,1);
    return 1;                // return mutated somewhere else
  } else {                
    next->haplotype[locus]=0;  
    if (next->desc_left!=NULL) {
	mut = recursemutaterecombsnp(next->desc_left,muttime,next->time
				     ,locus); 
	if (mut==1) {
	  mutatebelow(next->desc_right,locus,0);
	  return 1;
	} else {
	  mut = recursemutaterecombsnp(next->desc_right,muttime,next->time
				 ,locus);
	  return mut;
	}	
    } else return 0;
  }
}
/*************************************************************************/
void mutaterecombsnp(recombtree *rt)
{
  int locus,m;
  double length,muttime;

  for (locus=0;locus<rt->loci;locus++) {
  
    length = recomb_tree_length(rt->root[locus]->desc_left
				,rt->root[locus]->time) +
      recomb_tree_length(rt->root[locus]->desc_right,rt->root[locus]->time);
    muttime=unif_rand()*length;
    Rprintf("root time %g length %g, muttime %g\n",rt->root[locus]->time,length,muttime);
    m = recursemutaterecombsnp(rt->root[locus],&muttime,rt->root[locus]->time
			   ,locus); 
    if (m!=1) Rprintf("Error here at locus %d, we have %d\n",locus,m);
  }
}
/*************************************************************************/
void simulatelinkedSTR(int *n, int *loci, double *r,int *nr
			    , double *theta, int *data)
{
  recombtree rt;
  int i,j;
  
  rt=generaterecombtree(*n,*loci,r,*nr);
  mutaterecombSTR(&rt,theta);
  for (i=0;i<*loci;i++) {
    for (j=0;j<*n;j++) {
      data[i*(*n)+j]=rt.sample[j].haplotype[i];
    }
  }
  destroy_recombtree(&rt);
  return;
}
/*************************************************************************/
void simulatelinkedSNP(int *n, int *loci, double *r,int *nr, int *data)
{
  recombtree rt;
  int i,j;
  
  rt=generaterecombtree(*n,*loci,r,*nr);
  countleaves(&rt);
  mutaterecombsnp(&rt);
  for (i=0;i<*loci;i++) {
    for (j=0;j<*n;j++) {
      data[i*(*n)+j]=rt.sample[j].haplotype[i];
    }
  }
  destroy_recombtree(&rt);
  return;
}
/*************************************************************************/
void simulatetimeroot(int *n, int *loci, double *r, int *nr, double *troot)
{
  recombtree rt;
  int i;
  
  rt= generaterecombtree(*n,*loci,r,*nr);
  for (i=0;i<*loci;i++) troot[i]=rt.root[i]->time;
  destroy_recombtree(&rt);
  return;
}
/**************************************************************************/
recombtree generaterecombtree(int n, int loci, double *r, int nr)
{
  recombtree rt;
  int maxnodes,i,nleft,wh1,wh2,*nleftloc,a,maxancs,loc_left,lab;
  double rate[2],*recombrates,time=0.0;
  node **current,*newnode,*newnode2;
  GetRNGstate(); //set the R random generator
  
  rt.n=n;rt.loci=loci;
  maxnodes=4*n;maxancs=4*n;
  rt.sample = node_vector(n,loci);
  rt.root=(node**)Calloc(loci,node*);
  rt.ancestor=(node **)Calloc(maxancs,node*);
  initial_rates(n,loci,r,rate);
  current=(node **)Calloc(maxnodes,node*);
  for (i=0;i<n;i++) current[i]=&(rt.sample[i]);  // current active nodes
  nleftloc=(int*)Calloc(loci,int);               // number left at each locus
  for (i=0;i<loci;i++) nleftloc[i]=n;
  
  lab=n;a=0;loc_left=loci;nleft=n;
  for (;;) {
    if (a==maxancs-1||a==maxancs) {  // allocate more memory
      maxancs *=4;
      rt.ancestor=(node **)Realloc(rt.ancestor,maxancs,node *);
    }
    newnode=getnewnode(loci);           // get the next node (of either type)       
    rt.ancestor[a++]=newnode;           // and place it in the right position
    time+=exp_rand()/(rate[0]+rate[1]); // time until the next event
    newnode->time=time;
    if (unif_rand()<rate[0]/(rate[0]+rate[1])) { // coalescence
      wh1=(int)(unif_rand()*(double)nleft);
      wh2=(int)(unif_rand()*(double)(nleft-1));
      if (wh2>=wh1) wh2++;
      newnode->desc_left=current[wh1];
      newnode->desc_right=current[wh2];
      for (i=0;i<loci;i++) {
	if (current[wh1]->haplotype[i]>0||current[wh2]->haplotype[i]>0) {
	  newnode->haplotype[i]=1;
	  if (current[wh1]->haplotype[i]>0&&current[wh2]->haplotype[i]>0) {
	    nleftloc[i]-=1;
	    if (nleftloc[i]==1) { //root at position i
	      rt.root[i]=newnode;
	      loc_left-=1;
	      newnode->haplotype[i]=0; //not needed anymore ?
	    }
	  }
	}	
      }
      
      rate[1]-=diffrrate(current[wh1]->haplotype,current[wh2]->haplotype,
			 newnode->haplotype,loci,r);
      current[wh1]=newnode;
      current[wh2]=current[nleft-1];
      nleft-=1;
      rate[0] = (double)(nleft*(nleft-1))/2.0;
      if (loc_left==0) break;
    } else {     //recombination
      newnode2=getnewnode(loci);
      newnode2->time=time;
      rt.ancestor[a++]=newnode2;

      recombrates=get_recombrates(current,nleft,loci,r);
      wh1=gen_from_probs0(recombrates,nleft);
      Free(recombrates);
      wh2=which_recomb(current[wh1],loci,r);
      current[wh1]->recomb_pos=wh2;
      newnode->desc_left=current[wh1];
      newnode2->desc_left=current[wh1];
      for (i=0;i<=wh2;i++) 
	newnode->haplotype[i]=current[wh1]->haplotype[i];
      for (i=wh2+1;i<loci;i++)
	newnode2->haplotype[i]=current[wh1]->haplotype[i];
      rate[1]+=diffrrate(newnode2->haplotype,newnode->haplotype,
			 current[wh1]->haplotype,loci,r);
      
      if (nleft==maxnodes) {
	maxnodes*=2;
	current=(node**)Realloc(current,maxnodes,node *);
      } 
      current[wh1]=newnode;
      current[nleft]=newnode2;
      nleft+=1;
      rate[0] = (double)(nleft*(nleft-1))/2.0;
    }
  }
  rt.ancestor=(node **)Realloc(rt.ancestor,a,node *);
  rt.nancs=a;
  Rprintf("Ancestors = %d\n",a);
  Free(nleftloc);Free(current);
  PutRNGstate(); 
  return rt;
}
/*************************************************************************/
void destroy_recombtree(recombtree *t)
{
  int i;
  
  Free(t->root);
  for (i=0;i<t->nancs;i++) {
    Free(t->ancestor[i]->haplotype);
    Free(t->ancestor[i]);
  }
  Free(t->ancestor);
  for (i=0;i<t->n;i++) Free(t->sample[i].haplotype);
  Free(t->sample);
  t->n=0;
  t->loci=0;
  t->nancs=0;
}
/*************************************************************************/
node *node_vector(int n, int loci)
{
  int i,j;
  node *s;
  
  s=(node *)Calloc(n,node);
  for (i=0;i<n;i++) {
    s[i].desc_right=NULL; 
    s[i].desc_left=NULL;
    s[i].haplotype=(int*)Calloc(loci,int);
    s[i].leaf=i+1;
    for (j=0;j<loci;j++) s[i].haplotype[j]=1;
  }
  
  return s;
}
/***************************************************************************/
node *getnewnode(int loci)
{
  node *s;
  
  s=(node *)Calloc(1,node);
  s->haplotype=(int*)Calloc(loci,int);
  s->desc_right=NULL; s->desc_left=NULL;
  s->leaf=0;
  return s;
}
/***************************************************************************/
void initial_rates(int n, int loci, double *r,double rate[2])
{
  int i;
  rate[0]=(double)(n*(n-1))/2.;
  rate[1]=0.;
  for (i=0;i<loci-1;i++) rate[1]+=(double)n*r[i];
}
/***************************************************************************/
double *get_recombrates(node **current,int n, int loci,double *r)
{
  int i,j,k,l;
  double *rr;
	
  rr=(double *)Calloc(n,double);
  for (i=0;i<n;i++) {
    for (j=0;j<loci;j++) {   // get rid of those at the end that are not needed
      if (current[i]->haplotype[j]>0) break;
    }
    for (k=loci-1;k>0;k--) {
      if (current[i]->haplotype[k]>0) break;
    }
    if (j!=k) {              // and look at those remaining
      for (l=j;l<k;l++) rr[i] += r[l];
    }
  }
  return rr;
}
/****************************************************************************/
int which_recomb(node *any, int loci, double *r)
{
  int j,k;
	
  for (j=0;j<loci;j++) {
    if (any->haplotype[j]>0) break;
  }
  for (k=loci-1;k>0;k--) {
    if (any->haplotype[k]>0) break;
  }
  if (k<=j) {
    printf("something is going wrong here\n");
  }
  if (k-j==1) return j;
  //	Rprintf("j=%d k=%d\n",j,k);
  return j+gen_from_probs0(r+j,k-j); 
}
/****************************************************************************/
double diffrrate(int *h1, int *h2, int *h3,int loci, double *r)
{
  int j,k,l;
  double diff=0.0;

  for (j=0;j<loci;j++) {
    if (h1[j]>0) break;
  }
  for (k=loci-1;k>0;k--) {
    if (h1[k]>0) break;
  }
  for (l=j;l<k;l++) diff +=  r[l];
	
  for (j=0;j<loci;j++) {
    if (h2[j]>0) break;
  }
  for (k=loci-1;k>0;k--) {
    if (h2[k]>0) break;
  }
  for (l=j;l<k;l++) diff +=  r[l];

  for (j=0;j<loci;j++) {
    if (h3[j]>0) break;
  }
  for (k=loci-1;k>0;k--) {
    if (h3[k]>0) break;
  }
  for (l=j;l<k;l++) diff -=  r[l];
  return diff;
}
/***************************************************************************/




