/* a file of functions to load into R
   this file holds the functions for transferring
   data to and from R in the form of trees.   I also need
   ways to transfer data to and from R for linked data also
   I'll try to fix that
*/

#include <R.h>
#include <R_ext/PrtUtil.h>
#include <Rmath.h>
#include "treeutil.h"

#ifdef WIN32
#define DECLSPEC __declspec(dllexport) 
#else 
#define DECLSPEC 
#endif

DECLSPEC
void coalescentonepop(int *n, double *t, int *d1, int *d2);
DECLSPEC
void coalescentisland(int *n,int *npops, double *M,
			   double *t, int *d1, int *d2);
DECLSPEC
void coalescentsplit(int *n,int *npops, double *split,double *t, 
		     int *d1, int *d2);

DECLSPEC
void mutateSTR(int *n, double *t, 
		    int *d1, int *d2, int *dat, int *nSTR, double *theta);
DECLSPEC
void mutateK(int *n, double *t, 
		  int *d1, int *d2, int *dat, int *K, int *nK, double *theta);
DECLSPEC
void mutateinf(int *n, double *t, 
		    int *d1, int *d2, int *dat, int *ninf);
DECLSPEC
void meancoaltimewithinbetween(int *n, int *populations,int *npops
			       , double *t,int *d1, int *d2, double *res);
DECLSPEC
void meancoaltimewithin(int *n, int *populations,int *whpop, double *t, 
			     int *d1, int *d2, double *res);

/*******************************************************************/
DECLSPEC
     void coalescentonepop(int *n, double *t, int *d1, int *d2)
/*  need to pass the sample size n and the number of 
replicates needed and vectors of length reps*(n-1) for *t, *d1 and
*d2 */          
{
  int i,*labels,wh1,wh2;
  double height=0.0;

  GetRNGstate(); //set the R random generator
  labels=(int *)R_alloc(*n,sizeof(int));
  for (i=0;i< *n;i++) labels[i]=i+1;
  for (i=0;i< *n-1;i++) {
    height+=exp_rand()*2./(double)(( *n-i)*(*n-i-1));
    t[i]=height;
    wh1=(int)(unif_rand()*(*n-i));
    wh2=(int)(unif_rand()*(*n-i-1));
    if (wh2>=wh1) wh2++;
    d1[i]=labels[wh1];
    d2[i]=labels[wh2];
    labels[wh1]=*n+i+1;
    labels[wh2]=labels[*n-i-1];
  }
  PutRNGstate(); // kill the R generator
  return;
}
int getlabel(int pos,int *l, int *nl, int *d1, int *d2, int n) 
{
  if (pos==0) {
    Rprintf("error1a\n");
    return 0;
  }
  pos-=1;
  
  if (nl[pos]) return nl[pos];
  if (pos<n) {
    Rprintf("error should never get here..........pos %d nl %d l %d\n"
	    ,pos,nl[pos],l[pos]);
    return 0;
  }
  if(!l[pos]) {
    Rprintf("error should never get here......b\n");
    return 0;
  }
  pos -= n;
  if (l[d1[pos]-1]) return getlabel(d1[pos],l,nl,d1,d2,n);
  else if (l[d2[pos]-1]) return getlabel(d2[pos],l,nl,d1,d2,n);
  else 	{
    Rprintf("error should never get here.pos %d\n",pos);
    return 0;
  }
}
/*******************************************************************/
/*  take a tree defined by *t, *d1 and *d2, and subtract the 
    tree from the sample defined by the *n1 elements of label      
    the new tree is returned in *t, *d1 and *d2                    
    we need the variable label to be padded out to 2*n             */
DECLSPEC
void extracttree(int *label, int *n1, int *n,double *t, int *d1, int *d2)
{
  int i,count=0,*newlabel,pos,*nd1,*nd2;
  
  for (i=0;i<*n-1;i++) 
    label[*n+i]=label[d1[i]-1]||label[d2[i]-1];
  
  /* get the new labellings for the nodes that are present  */;
  newlabel= (int *)R_alloc(*n*2,sizeof(int));
  nd1= (int *)R_alloc(*n1,sizeof(int));
  nd2= (int *)R_alloc(*n1,sizeof(int));
  for (count=0,i=0;i<*n;i++) {
    if (label[i]) newlabel[i]=++count;
    else newlabel[i]=0;
  }
  for (i=0;i<*n-1;i++) {
    if (label[d1[i]-1]&&label[d2[i]-1]) {
      newlabel[*n+i]=++count;
    } else newlabel[*n+i]=0;
  }
  for (pos=0,i=0;i<*n-1;i++) {
    if (label[d1[i]-1]&&label[d2[i]-1]) {
      t[pos]=t[i];
      nd1[pos]=getlabel(d1[i],label,newlabel,d1,d2,*n);
      nd2[pos]=getlabel(d2[i],label,newlabel,d1,d2,*n);
      pos+=1;
    } 
  } 
  for (i=0;i<*n1-1;i++) {
    d1[i]=nd1[i];
    d2[i]=nd2[i];
  }
  return;
}
/******************************************************************/
void calc_rate(double *rate, int *n, int npops, double M)
{
  int i;
  rate[0]=0.0;
  for (i=0;i<npops;i++) {
    if (n[i]>0) {
      if (n[i]==1) {
	rate[i+1] = 0; 
	rate[npops+i+1]=M;
	rate[0]+=M;
      } else {
	rate[i+1]= (double)(n[i]*(n[i]-1)) /2.;	
	rate[npops+i+1] = (double)(n[i])*M;
	rate[0] += rate[i+1]+rate[npops+i+1];
      }
    } else {
      rate[i+1] = 0.0;
      rate[npops+i+1]=0.0;
    }
  }
  return;
}
/******************************************************************/
DECLSPEC
void coalescentisland(int *n,int *npops, double *M,double *t, int *d1, int *d2)
/*  need to pass the sample size n, the populations npops the migration 
    rate M and vectors of length (totn-1) for *t, *d1 and *d2 */          
{
    int i,j,**labels,which,wh1,wh2,tot=0,count=1;
    double height=0.0,*rate;
	
    for (i=0;i<*npops;i++) tot+=n[i];
    
    GetRNGstate(); //set the R random generator
    // Set up the various arrays
    rate=(double *)R_alloc(2*(*npops)+1,sizeof(double));
    labels=(int **)R_alloc(*npops,sizeof(int*));
    for (i=0;i< (*npops);i++) {
      labels[i]=(int *)R_alloc(tot,sizeof(int));
      for (j=0;j<n[i];j++) labels[i][j] = count++;
    }
    for (i=0;i< tot-1;) {
      calc_rate(rate,n,*npops,*M);
      height+=exp_rand()/rate[0];
      which=gen_from_probs(rate,2*(*npops));
      if (which<=*npops) {
	which-=1;
	t[i]=height;
	wh1=(int)(unif_rand()*(double)n[which]);
	wh2=(int)(unif_rand()*(double)n[which]-1.);
	if (wh2>=wh1) wh2++;
	d1[i]=labels[which][wh1];
	d2[i]=labels[which][wh2];
	labels[which][wh1]=tot+i+1;
	labels[which][wh2]=labels[which][n[which]-1];
	n[which]-=1;
	i++;
      } else {
	which = which -1 - *npops;
	wh1=(int)(unif_rand()*(double)n[which]);
	wh2=(int)(unif_rand()*(double)(*npops-2));
	if (wh2>=which) wh2++;
	labels[wh2][n[wh2]]=labels[which][wh1];
	n[wh2]+=1;
	labels[which][wh1]=labels[which][n[which]-1];
	n[which]-=1;
      }
    }
    PutRNGstate(); // kill the R generator
    return;
}
/******************************************************************/
double nexttime(int n)
{
  return exp_rand()*2./(double)(n*(n-1.));
}
/******************************************************************/
DECLSPEC
void coalescentsplit(int *n,int *npops, double *split,double *t
		     , int *d1, int *d2)
/*  need to pass the sample size n, the populations npops the migration 
    rate M and vectors of length (totn-1) for *t, *d1 and *d2 */          
{
    int i,j,**labels,wh1,wh2,tot=0,count=1,where=0,left;
    double height=0.0;
	
    for (i=0;i<*npops;i++) tot+=n[i];  // get the total sample size
    
    GetRNGstate(); //set the R random generator
    // Set up the labels
    labels=(int **)R_alloc(*npops,sizeof(int*));
    for (i=0;i< (*npops);i++) {
      labels[i]=(int *)R_alloc(tot,sizeof(int));
      for (j=0;j<n[i];j++) labels[i][j] = count++;
    }
    //  Do mini one population models for each population
    for (i=0;i<*npops;i++) {
      height=0.0;
      count=n[i];
      for (j=0;j<n[i]-1;j++) {
	height+=nexttime(count);
	if (height> *split) break;
	t[where]=height;
	wh1=(int)(unif_rand()*(double)count);
	wh2=(int)(unif_rand()*(double)count-1.);
	if (wh2>=wh1) wh2++;
	d1[where]=labels[i][wh1];
	d2[where]=labels[i][wh2];
	labels[i][wh1]=tot+where+1;
	labels[i][wh2]=labels[i][count-1];
	count -=1;
	where +=1;
      }
      n[i]=count;
    }
    // now gather together the rest and then finish off 
    left=n[0];count=n[0];
    for (i=1;i<*npops;i++) {
      left+=n[i];  // count up the rest
      for (j=0;j<n[i];j++) labels[0][count++]=labels[i][j];
    } 
    height=*split;
    count=left;
    for (i=0;i<left-1;i++) { 
      height+=nexttime(count);
      t[where]=height;
      wh1=(int)(unif_rand()*(count));
      wh2=(int)(unif_rand()*(count-1));
      if (wh2>=wh1) wh2++;
      d1[where]=labels[0][wh1];
      d2[where]=labels[0][wh2];
      labels[0][wh1]=tot+where+1;
      labels[0][wh2]=labels[0][count-1];
      where +=1;
      count-=1;
    }
  
    PutRNGstate(); // kill the R generator
    return;
}
/***************************************************************************/
/*   Generate the next time                                                */
/***************************************************************************/
double getnexttimeprop(double last,double left,int gtype,double *a, double pr)
{
  double t;
  if (gtype==0) {
    return last+exp_rand()*2.*pr/(double)(left*(left-1.));
  } else if (gtype==1) {
    if (fabs(a[0])>1E-3) {
      return log(
		 exp(a[0]*last)-log(unif_rand())*2.*a[0]*pr/
		 (left*(left-1.0)) )/a[0] -last;
    }
    else {
      return -log(unif_rand())*2.0*pr/(left*(left-1.0));
    }
  } else if (gtype==2) {
    if (last>=a[1])
      return -log(unif_rand())*2.0*pr/(left*(left-1.0));
    else {
      t = log(
	      exp(a[0]*last)-log(unif_rand())*exp(a[0]*a[1])*2.*a[0]*pr
	      /(left*(left-1.0)))/a[0]-last;
      if (t+last> a[1]) 
	return a[1] + getnexttimeprop(a[1],left,gtype,a,pr);
      else return t;
    }	
  } else {
    Rprintf("not written yet in getnexttimechange");
    exit(-1);
    return 0.0;
  }
}
/******************************************************************/
DECLSPEC
void coalescentgrowthsplit(int *n,int *npops, double *split,int *gtype
			   ,double *growthpars,double *t, int *d1, int *d2)
/*  need to pass the sample size n, the populations npops the migration 
    rate M and vectors of length (totn-1) for *t, *d1 and *d2 */          
{
    int i,j,**labels,wh1,wh2,tot=0,count=1,where=0,left;
    double height=0.0;
	
    for (i=0;i<*npops;i++) tot+=n[i];  // get the total sample size
    
    GetRNGstate(); //set the R random generator
    // Set up the labels
    labels=(int **)R_alloc(*npops,sizeof(int*));
    for (i=0;i< (*npops);i++) {
      labels[i]=(int *)R_alloc(tot,sizeof(int));
      for (j=0;j<n[i];j++) labels[i][j] = count++;
    }
    //  Do mini one population models for each population
    for (i=0;i<*npops;i++) {
      height=0.0;
      count=n[i];
      for (j=0;j<n[i]-1;j++) {
	height=getnexttimeprop(height,count,*gtype,growthpars,1/(double)(*npops));
	if (height> *split) break;
	t[where]=height;
	wh1=(int)(unif_rand()*(double)count);
	wh2=(int)(unif_rand()*(double)count-1.);
	if (wh2>=wh1) wh2++;
	d1[where]=labels[i][wh1];
	d2[where]=labels[i][wh2];
	labels[i][wh1]=tot+where+1;
	labels[i][wh2]=labels[i][count-1];
	count -=1;
	where +=1;
      }
      n[i]=count;
    }
    // now gather together the rest and then finish off 
    left=n[0];count=n[0];
    for (i=1;i<*npops;i++) {
      left+=n[i];  // count up the rest
      for (j=0;j<n[i];j++) labels[0][count++]=labels[i][j];
    } 
    height=*split;
    count=left;
    for (i=0;i<left-1;i++) { 
      height=getnexttimeprop(height,count,*gtype,growthpars,1.0);
      t[where]=height;
      wh1=(int)(unif_rand()*(count));
      wh2=(int)(unif_rand()*(count-1));
      if (wh2>=wh1) wh2++;
      d1[where]=labels[0][wh1];
      d2[where]=labels[0][wh2];
      labels[0][wh1]=tot+where+1;
      labels[0][wh2]=labels[0][count-1];
      where +=1;
      count-=1;
    }
  
    PutRNGstate(); // kill the R generator
    return;
}
/******************************************************************/
DECLSPEC
void mutateSTR(int *n, double *t, int *d1, int *d2, int *dat, 
	       int *nSTR, double *theta)
     /*  note that we need some extra space in the simulation in that 
	 the length of dat should have an extra spare length of n-1    */ 
{
  int locus,k,*data;
  double thet;

  GetRNGstate();

  for (locus=0;locus< *nSTR;locus++) {
    thet=theta[locus]*0.25;
    data=dat+locus*(*n);
    data[2*(*n)-2]=0;
    for (k= *n-2;k>=0;k--) {
      if (d1[k]<*n) {
	data[d1[k]-1] = data[*n+k] +
	  (int)rpois(thet*t[k])-(int)rpois(thet*t[k]);
      } else {
	data[d1[k]-1]=data[*n+k] +
	  (int)rpois(thet*(t[k]-t[d1[k]-1-*n]))
	  -(int)rpois(thet*(t[k]-t[d1[k]-1-*n]));
      }
      if (d2[k]<*n) {
	data[d2[k]-1]=data[*n+k] + 
	  (int)rpois(thet*t[k])
	  -(int)rpois(thet*t[k]);
      } else {
	data[d2[k]-1]=data[*n+k] + 
	  (int)rpois(thet*(t[k]-t[d2[k]-1-*n]))
	  -(int)rpois(thet*(t[k]-t[d2[k]-1-*n]));
      }	
    }
  }
  PutRNGstate();
  return;
}
/******************************************************************/
DECLSPEC void mutateK(int *n, double *t,  int *d1, int *d2, int *dat, 
		      int *K, int *nK, double *theta)
     /*  note that we need some extra space in the simulation in that 
	 the length of dat should have an extra spare length of n-1    */ 
{
  int locus,k,*data;
  double thet;

  GetRNGstate();

  for (locus=0;locus< *nK;locus++) {
    thet=theta[locus]*0.5;
    data=dat+locus*(*n);
    data[2*(*n)-2]=1+(int)(unif_rand()*(double)(*K));
    for (k= *n-2;k>=0;k--) {
      if (d1[k]<*n) {
	if (unif_rand()<thet*t[k])
	  data[d1[k]-1] = 1+(int)(unif_rand()*(double)(*K));
	else data[d1[k]-1]=data[*n+k];
      } else {
	if (unif_rand()<thet*(t[k]-t[d1[k]-1-*n]))
	  data[d1[k]-1] = 1+(int)(unif_rand()*(double)(*K));
	else  data[d1[k]-1]=data[*n+k];
      }
      if (d2[k]<*n) {
	if (unif_rand()<thet*t[k])
	  data[d2[k]-1] = 1+(int)(unif_rand()*(double)(*K));
	else data[d2[k]-1]=data[*n+k];
      } else {
	if (unif_rand()<thet*(t[k]-t[d2[k]-1-*n]))
	  data[d2[k]-1] = 1+(int)(unif_rand()*(double)(*K));
	else  data[d2[k]-1]=data[*n+k];
      }
    }
  }
  PutRNGstate();
  return;
}
/******************************************************************/
DECLSPEC void meancoaltime(int *n, double *t, 
			   int *d1, int *d2, double *res)
{
  int i,*count;
  double sm=0.0;

  count=(int *)R_alloc(2*(*n),sizeof(int));
  for (i=1;i<= *n;i++) count[i]=1;

  for (i=0;i< *n-1;i++) {
    sm += count[d1[i]]*count[d2[i]]*t[i];
    count[*n+1+i]=count[d1[i]]+count[d2[i]];
  }
  *res = 2.*sm / ((*n)*(*n-1));
  return;
}
/******************************************************************/
void getlengths(double *l1, double *l2, int n, double *t, int *d1, int *d2)
{
  int i;
  
  for (i=0;i<n-1;i++) {
    if (d1[i]<= n) {
      l1[i]=t[i];
    } else {
      l1[i]=t[i]-t[d1[i]- 1 - n];
    }
    if (d2[i]<= n) {
      l2[i]=t[i];
    } else {
      l2[i]=t[i]-t[d2[i]- 1 - n];
    }
  }
  return;	
}
/******************************************************************/
void change_below(int *data, int *da, int *db, int label, int n) 
{
  if (label<=n) { 
    data[label-1]=1;
  } else {
    change_below(data,da,db,da[label-n-1],n);
    change_below(data,da,db,db[label-n-1],n);
  }
}
/******************************************************************/
DECLSPEC
     void mutateinf(int *n, double *t, int *d1, int *d2, int *dat, int *ninf)
{
  int locus,i,where,below;
  double *l;
  
  GetRNGstate();
  l=(double *)R_alloc(2*(*n)-1,sizeof(double));
  getlengths(l+1,l+*n,*n,t,d1,d2);
  for (locus=0;locus< *ninf;locus++) {
    for (i=0;i< *n;i++) dat[locus*(*n)+i]=0;  
    where=gen_from_probs(l,2*(*n-1));
    if (where <= *n-1) below = d1[where-1];
    else below = d2[where - *n];
    change_below(dat+(*n)*locus,d1,d2,below,*n);
  }
  PutRNGstate();
  return;
}
/******************************************************************/
DECLSPEC
void meancoaltimewithin(int *n, int *populations,int *whpop, double *t, 
			     int *d1, int *d2, double *res)
{
  int i,*count,tot=0;
  double sm=0.0;

  count=(int *)R_alloc(2*(*n),sizeof(int));
  for (i=1;i<= *n;i++) {
    if (populations[i]==*whpop) {
      count[i]=1;
      tot +=1;
    } else count[i]=0;
  }
  for (i=0;i< *n-1;i++) {
    sm += count[d1[i]]*count[d2[i]]*t[i];
    count[*n+1+i]=count[d1[i]]+count[d2[i]];
  }
  *res = 2.*sm / ((tot)*(tot-1));
  return;
}
/******************************************************************/
DECLSPEC
void meancoaltimewithinbetween(int *n, int *populations,int *npops , 
			       double *t,int *d1, int *d2, double *res)
{
  int i,j,k,**count,nw=0,nb=0;
  double smw=0.0,smb=0.0;

  count = (int **)R_alloc(*npops+1,sizeof(int*));
  for (i=1;i<=*npops;i++) {
    count[i]=(int *)R_alloc(2*(*n),sizeof(int));
    for (j=1;j<2*(*n);j++) count[i][j]=0;
  }
  for (i=1;i<= *n;i++) count[populations[i-1]][i]=1;
	
  for (i=0;i< *n-1;i++) {
    for (j=1;j<=*npops;j++) {
      for (k=1;k<=*npops;k++) {
	if (j==k) {
	  smw += count[j][d1[i]]*count[j][d2[i]]*t[i];
	  nw += count[j][d1[i]]*count[j][d2[i]];
	} else {
	  smb += count[j][d1[i]]*count[k][d2[i]]*t[i];
	  nb += count[j][d1[i]]*count[k][d2[i]];
	}
      }
      count[j][*n+1+i]=count[j][d1[i]]+count[j][d2[i]];
    }
  }
  res[0] = smw / nw;
  res[1] = smb/nb;
  return;
}
/******************************************************************/
